--[[********************************************************************
created:	2009/06/18
filename: 	AStar.lua
author:		Daniel Wellman

purpose:	A fun 'game' of cat and mouse where monsters chase you!

License: This software may be reproduced and distributed free of charge as
as long as credit is given to the author. The author assumes no
responsibility. The software is provided as is.
*********************************************************************]]

AStar = {}

--A* consumes a lot of time, let's do some micro-optimizations!
--Bind commonly used functions to Lua's virtual registers
--The function lookup is about 30% faster over global calls
local table_sort = table.sort
local math_abs = math.abs
local pairs = pairs
local table_insert = table.insert

local grid
local SetEntityLocation, GetEntityLocation
--Constants
local NUM_MONSTERS = 5 --Can have more (or less), just starting to run out of colors :{


do--Initialize
	local options = {
		Design = "Messy",
		DesignOptions = {"Messy", "Blizzard", "Empty",},
	}
	local function GetSpecializedGrid(width, height)
		--Specialized grid
		local grid = {}
		for x=1, width do
			grid[x] = {}
			for y=1, height do
				grid[x][y] = {X = x, Y = y}
			end
		end
		--Iterate each cell
		function grid:Iterate()
			local x, y = 0, 1
			return function()
				x = x + 1
				if x > width then
					x = 1
					y = y + 1
					if y > height then return end
				end
				return self[x][y]
			end
		end
		do--IterateNeighbors
			--Temp table, usually we can let Lua GC it
			--However, re-using small special purpose tables can greatly reduce garbage
			local t = {}
			local function AddNeighbor(self, x, y)
				if x <= width and x > 0 and y <= height and y > 0 then
					t[#t+1] = self[x][y]
				end
			end
			--Returns an iterator that iterates through the neighbors around this cell
			function grid:IterateNeighbors(cell)
				--Clear out the temp table
				for i=1, #t do t[i] = nil end
				local x, y = cell.X, cell.Y
				AddNeighbor(self, x-1, y)
				AddNeighbor(self, x+1, y)
				AddNeighbor(self, x, y+1)
				AddNeighbor(self, x, y-1)
				--Diagonals
				AddNeighbor(self, x-1, y-1)
				AddNeighbor(self, x+1, y-1)
				AddNeighbor(self, x-1, y+1)
				AddNeighbor(self, x+1, y+1)
				return ipairs(t)
			end
		end
		return grid
	end
	function AStar:Initialize(width, height)
		grid = GetSpecializedGrid(width, height)

		for x, y in Designs:Iterate(options.Design) do
			grid[x][y].Wall = true
		end

		SetEntityLocation(1, 1, "You")
		SetEntityLocation(width, height, "Monster1")
		SetEntityLocation(1, height, "Monster2")
		SetEntityLocation(width, 1, "Monster3")
		SetEntityLocation(width - 5, 1, "Monster4")
		SetEntityLocation(1, height - 5, "Monster5")

		self.Lost = false
		self.Resettable = false
		self.IsInitialized = true
	end

	--Register with the menu
	Menu:RegisterSimulator("AStar", AStar, options)
end

--Description and mouse button function
function AStar:GetDetails()
	return "An A* path finding demo. You are * and $ are monsters. Run!", 
		"Hold for character to path, monsters will start to chase", 
		"Adds/removes walls"
end

do--Handles setting and removing our special entities
	local entities = {}
	function SetEntityLocation(x, y, which)
		if entities[which] then
			grid[entities[which].X][entities[which].Y][which] = false
		else
			entities[which] = {}
		end
		entities[which].X = x
		entities[which].Y = y
		grid[x][y][which] = true
	end
	function GetEntityLocation(which)
		return grid[entities[which].X][entities[which].Y]
	end
end

--Defined later, finds a path using A*
local FindPath
do--Update
	local HandleInput
	do
		local wallMode
		function HandleInput()
			if IsRightMousePressed() then
				local x, y = GetCursorPosition()
				if wallMode == nil then
					wallMode = not grid[x][y].Wall
				end
				grid[x][y].Wall = wallMode
			else
				wallMode = nil
			end
			if IsLeftMousePressed() then
				local x, y = GetCursorPosition()
				return true, grid[x][y]
			end
		end

	end
	local function CheckGameOver(monster, you)
		you = you or GetEntityLocation("You")
		if monster then
			return you.X == monster.X and you.Y == monster.Y
		else
			for i=1, NUM_MONSTERS do
				--Recursively determine if monster has killed the player!
				if CheckGameOver(GetEntityLocation("Monster" .. i), you) then
					return true
				end
			end
		end
		
	end
	local updateMonster
	local monsterNum = 1
	function AStar:Update(frame)
		if self.Lost then
			if not self.Resettable then
				self.Resettable = not IsLeftMousePressed()
			--Reset sim
			elseif IsLeftMouseClicked() then
				SetSimulator(self)
			end
			return
		end
		for cell in grid:Iterate() do
			--Reset cells, only one path can be found per update because of this
			cell.Closed = false
			cell.Open = false
		end
		local shouldMove, mouseCell = HandleInput()
		--Monster moves at a rate of once every NUM_MONSTERS frames
		if updateMonster then
			updateMonster = false
			monsterNum = monsterNum + 1
			if monsterNum > NUM_MONSTERS then
				monsterNum = 1
			end
			local monsterName = "Monster" .. monsterNum
			local nextCell = FindPath(GetEntityLocation(monsterName), GetEntityLocation("You"))
			if nextCell and nextCell ~= currentCell then
				SetEntityLocation(nextCell.X, nextCell.Y, monsterName)
				if CheckGameOver(nextCell, nil) then
					self.Lost = true
				end
				
			end
		--Left mouse is down, start simulating
		elseif shouldMove then
			local currentCell = GetEntityLocation("You")
			local nextCell = FindPath(currentCell, mouseCell)
			if nextCell and nextCell ~= currentCell then
				SetEntityLocation(nextCell.X, nextCell.Y, "You")
				updateMonster = true
				if CheckGameOver(nil, nextCell) then
					self.Lost = true
				end
			end

		end
		
	end
end

do--Drawing
	local function CellToColor(cell)
		if cell.Wall then
			return ' ', Colors.Black, Colors.Gold
		elseif cell.You then
			return '*', Colors.Green, Colors.Black
		elseif cell.Monster1 then
			return '$', Colors.Red, Colors.Black
		elseif cell.Monster2 then
			return '$', Colors.Purple, Colors.Black
		elseif cell.Monster3 then
			return '$', Colors.Yellow, Colors.Black
		elseif cell.Monster4 then
			return '$', Colors.White, Colors.Black
		elseif cell.Monster5 then
			return '$', Colors.Teal, Colors.Black
		end
		return ' ', Colors.Black, Colors.Black
	end
	function AStar:Draw()
		for cell in grid:Iterate() do
			DrawSquare(cell.X, cell.Y, CellToColor(cell))
		end
		if self.Lost then
			for x, y in Designs:Iterate("YouAreDead") do
				DrawSquare(x, y, '*', Colors.Green, Colors.Blue)
			end
		end
	end
end

--Actual A* implementation
do
	local openList = {}
	local startCell
	local endCell
	local closestCell 
	local function G(cell)
		local parent = cell.Parent
		if parent.X ~= cell.X and parent.Y ~= cell.Y then
			--Diagonal
			return 14 + parent.G
		else
			--Orthogonal
			return 10 + parent.G
		end
	end
	local function UpdateValues(cell, parent, forceUpdate)
		cell.Parent = parent
		--Inline H() for speed, manhattan distance
		cell.H = (math_abs(cell.X - endCell.X) + math_abs(cell.Y - endCell.Y)) * 10
		cell.G = G(cell)
		cell.F = cell.H + cell.G
		--The open list is a sorted array
		--However, table.sort is VERY slow and more than a few dozen cells
		--will bring the program to its knees
		--Instead, manually sort the array as cells are added
		if not cell.Open and not cell.Closed then
			--Empty, add it to the first index, or if the cell is >= just add it to the end
			if #openList == 0 or openList[#openList].F >= cell.F then
				openList[#openList + 1] = cell
			else
				--Find the slot to insert the cell
				local f = cell.F
				for i=1, #openList do
					if openList[i].F < f then
						table_insert(openList, i, cell)
						break
					end
				end
			end
			cell.Open = true
		end
	end
	--Pops the top of the sorted list
	local function GetClosest()
		local closestCell = openList[#openList]
		openList[#openList] = nil
		closestCell.Closed = true
		closestCell.Open = false
		return closestCell
	end
	local function IsWalkable(cell, neighbor)
		--We can't walk through walls, can we?
		if neighbor.Closed or neighbor.Wall then
			return false
		end
		--Don't allow diagonal if there's a block in the way
		if neighbor.X ~= cell.X and neighbor.Y ~= cell.Y then
			return not grid[neighbor.X][cell.Y].Wall and not grid[cell.X][neighbor.Y].Wall
		end
		return true
		
	end
	--Avoid making a bunch of garbage, table.sort is slow enough already
	local function Compare(A, B)
		return A.F > B.F
	end
	--Finds the next cell
	local function UpdateOpenList()
		closestCell = GetClosest()
		closestCell.Open = false
		closestCell.Closed = true
		for i, neighbor in grid:IterateNeighbors(closestCell) do
			if IsWalkable(closestCell, neighbor) then
				if neighbor.Open then
					--Check this neighbor to see if its shorter than our 'closest'
					if neighbor.G < closestCell.G then
						UpdateValues(neighbor, closestCell)
						--Should only rarely be needed
						--Re-sort the list so the new cell as at the top
						table_sort(openList, Compare)
					end
				else
					UpdateValues(neighbor, closestCell)
				end
			end

		end
	end
	--Finds the path and returns the next cell to move to
	function FindPath(startingCell, endingCell)
		for i=#openList, 1, -1 do openList[i] = nil end
		startCell = startingCell
		endCell = endingCell

		startCell.G = 0
		UpdateValues(startCell, startCell)

		while not endCell.Closed and #openList > 0 do
			UpdateOpenList()
		end
		if endCell.Closed then --We found a path
			local currentCell = closestCell
			--We want the next cell from our starting point
			while currentCell.Parent ~= startCell do
				currentCell = currentCell.Parent
			end
			return currentCell
		end
	end
end