--[[********************************************************************
created:	2009/10/10
filename: 	Beach.lua
author:		Daniel Wellman

purpose:	A relaxing simulation, watch the beach crash into the Blizzard beach

License: This software may be reproduced and distributed free of charge as
as long as credit is given to the author. The author assumes no
responsibility. The software is provided as is.
*********************************************************************]]

--Constants
local FLOW = 1.5
local EMITTER_PRESSURE = 10000
local EMITTER_DELAY = 15

Beach = {}

local emitters
local grid
do--Initialize
	function Beach:Initialize(width, height)
		local properties = {
			"Inert", 
			"Pressure",
			"NewPressure",
			"LastTurn",
		}
		grid = Grid:New(width, height, properties)

		--Initialize default cell values
		for x, y in grid:Iterate() do
			grid:Set(x, y, "Pressure", 0)
			grid:Set(x, y, "NewPressure", 0)
			grid:Set(x, y, "LastTurn", false)
		end

		--Insert the design
		for x, y in Designs:Iterate("Blizzard") do
			grid:Set(x, y, "Inert", true)
		end

		--Initialize emitters
		emitters = {}
		for x=1, WIDTH, 2 do
			emitters[x] = height
		end
	end
	--Add beach to the menu
	Menu:RegisterSimulator("Beach", Beach)

	--Returns the description and mouse controls
	function Beach:GetDetails()
		return "A relaxing simulation of waves crashing onto...a beach?", 
			"Nothing! Sit back and enjoy...order another pina colada", 
			"Also nothing! Call me up I'll make WOOSH sounds for you"
	end
end



do--Update
	local currentTurn = false
	--The grid stores the last two states of a cell using this method
	--Instead of duplicating the entire grid just store the last updates pressure
	--On the new frame move the pressure over, so we'll be one frame behind, but its not noticeable
	local function UpdateTurn(x, y)
		if grid(x, y, "LastTurn") ~= currentTurn then
			grid:Set(x, y, "LastTurn", currentTurn)
			grid:Set(x, y, "Pressure", grid(x, y, "NewPressure"))
		end
	end
	--Clamps a value between two other values
	local function Clamp(value, max, min)
		if value > max then
			return max
		end
		if value < min then
			return min
		end
		return value
	end
	--Linear interpolation between values
	local function Lerp(value1, value2, amount)
		return value1 + ((value2 - value1) * amount)
	end
	local emitterDelay = 0
	--Handles the water emitters that makes it look like a wave crashing onto the beach
	local function UpdateEmitters(dt)
		if emitterDelay > 0 then
			emitterDelay = emitterDelay - dt
		else
			emitterDelay = EMITTER_DELAY
			for x, y in pairs(emitters) do
				grid:Set(x, y, "NewPressure", grid:Get(x, y, "NewPressure") + EMITTER_PRESSURE)
			end
		end
	end
	function Beach:Update(dt)
		UpdateEmitters(dt)

		--Usually keep locals as close to their definition as possible
		--However, with many cells this micro-optimization results in noticeable speedup
		local pressureDelta, flow
		local pressure, neighPressure
		for x, y in grid:Iterate() do
			if not grid(x, y, "Inert") then
				UpdateTurn(x, y)
				pressure = grid(x, y, "Pressure")
				--Some water is seeping into the sand
				grid:Set(x, y, "NewPressure", Lerp(grid(x, y, "NewPressure"), 0, dt * .5))
				for nx, ny in grid:IterateNeighbors(x, y) do
					if not grid(nx, ny, "Inert") then
						UpdateTurn(nx, ny)
						neighPressure = grid(nx, ny, "Pressure") 
						pressureDelta = pressure - neighPressure
						flow = pressureDelta * FLOW * dt
						flow = Clamp(flow, pressure / 4, -neighPressure / 4)

						grid:Set(x, y, "NewPressure", grid(x, y, "NewPressure") - flow)
						grid:Set(nx, ny, "NewPressure", grid(nx, ny, "NewPressure") + flow)
					end

				end
			end
		end
		currentTurn = not currentTurn
	end
end

local GetSand
do
	local chars = {}
	function GetSand(x, y)
		if not chars[x * y] then
			local possibles = "  . ',:"
			chars[x * y] = string.sub(possibles, math.random(#possibles))
		end
		return chars[x * y]
	end
end
do--Drawing
	local function CellToColor(x, y)
		if grid(x, y, "Inert") then
			return GetSand(x, y), Colors.Gold, Colors.Yellow
		end
		local pressure = grid(x, y, "Pressure")
		if pressure > 100 then
			return ' ', Colors.DarkBlue, Colors.DarkBlue
		elseif pressure > 45 then
			return '*', Colors.Blue, Colors.DarkBlue
		elseif pressure > 35 then
			return ' ', Colors.Blue, Colors.Blue
		elseif pressure > 25 then
			return '*', Colors.DarkTeal, Colors.Blue
		elseif pressure > 15 then
			return ' ', Colors.DarkTeal, Colors.DarkTeal
		elseif pressure > 8 then
			return '*', Colors.Teal, Colors.DarkTeal
		elseif pressure > 4 then
			return ' ', Colors.Teal, Colors.Teal
		elseif pressure > 1 then
			return '*', Colors.Teal, Colors.Black
		end
		return ' ', Colors.White, Colors.Black
	end
	function Beach:Draw()
		for x, y in grid:Iterate() do
			DrawSquare(x, y, CellToColor(x, y))
		end
	end
end