--[[********************************************************************
created:	2009/10/10
filename: 	Cells.lua
author:		Daniel Wellman

purpose:	Conway's Game of Life

License: This software may be reproduced and distributed free of charge as
as long as credit is given to the author. The author assumes no
responsibility. The software is provided as is.
*********************************************************************]]

--Constants
local TIME_TO_WAIT_AFTER_INPUT = 2

Cells = {}

local grid
do
	local options = {
		Design = "ExploderFlower",
		DesignOptions = {"ExploderFlower", "Exploder", "Glider", "Flier", "Blizzard", "GliderGun", "Empty"},
		Speed = "Medium",
		SpeedOptions = {"Slow", "Medium", "Fast", "Unlimited"},
	}
	function Cells:Initialize(width, height)
		grid = Grid:New(width, height, {"Alive", "WasAlive"})
		
		if options.Design ~= "Empty" then
			for x, y in Designs:Iterate(options.Design) do
				grid:Set(x, y, "Alive", true)
			end
		end

		if options.Speed == "Unlimited" then
			self.UpdateRate = 0
		elseif options.Speed == "Fast" then
			self.UpdateRate = .05
		elseif options.Speed == "Medium" then
			self.UpdateRate = .15
		elseif options.Speed == "Slow" then
			self.UpdateRate = .35
		end
	end

	--Register with the menu system
	Menu:RegisterSimulator("Cells", Cells, options)

	--Returns the description and mouse controls
	function Cells:GetDetails()
		return "Conway's Game of Life written in Lua.", 
			"Create a cell", 
			"Destroy a cell"
	end
end

local HandleInput
do
	local inputDelay = 0
	--Set cells under the left mouse to alive, and kill the ones under the right mouse
	function IsHandlingInput(dt)
		if IsLeftMousePressed() then
			local x, y = GetCursorPosition()
			grid:Set(x, y, "Alive", true)
			inputDelay = TIME_TO_WAIT_AFTER_INPUT
		end
		if IsRightMousePressed() then
			local x, y = GetCursorPosition()
			grid:Set(x, y, "Alive", false)
			inputDelay = TIME_TO_WAIT_AFTER_INPUT
		end

		--Delay if a button was recently pressed
		if inputDelay > 0 then
			inputDelay = inputDelay - dt
			return true
		end
		return false
	end
end

do
	local updateDelay = 0
	function Cells:Update(dt)
		--Only update if no input is being done from the user
		if not IsHandlingInput(dt) then
			--Delay between updates or the simulation ends immediately
			if updateDelay > 0 then
				updateDelay = updateDelay - dt
			else
				updateDelay = updateDelay + self.UpdateRate
				--Save last alive state to avoid double buffering
				for x, y in grid:Iterate() do
					grid:Set(x, y, "WasAlive", grid(x, y, "Alive"))
				end
				for x, y in grid:Iterate() do
					--Cells die or are born depending on their number of neighbors
					local numNeighbors = grid:GetNumNeighbors(x, y, "WasAlive")
					if grid(x, y, "Alive") then
						if numNeighbors ~= 2 and numNeighbors ~= 3 then
							grid:Set(x, y, "Alive", false)
						end
					else
						if numNeighbors == 3 then
							grid:Set(x, y, "Alive", true)
						end
					end
				end
			end
		end
	end
end

function Cells:Draw()
	Clear()
	for x, y in grid:Iterate() do
		if grid(x, y, "Alive") then
			DrawSquare(x, y, 'O', Colors.White, Colors.Black)
		end
	end
end