--[[********************************************************************
created:	2009/10/10
filename: 	Fire.lua
author:		Daniel Wellman

purpose:	Simulates fire spreading by air temperature using Cellular Automata

License: This software may be reproduced and distributed free of charge as
as long as credit is given to the author. The author assumes no
responsibility. The software is provided as is.
*********************************************************************]]

Fire = {}

local grid
local TreeProperties = {
	--Rate at which 'health' is consumed
	BurnRate = .05,
	--The max temperature of the burning cell
	MaxBurn = 4,
	--The min temperature before the object will start burning
	FlashPoint = 1,
	--The temperature that the object emits while burning
	BurnTemp = 10,
}
local BarrelProperties = {
	--Rate at which 'health' is consumed
	BurnRate = 50,
	--The max temperature of the burning cell
	MaxBurn = 500,
	--The min temperature before the object will start burning
	FlashPoint = 1,
	--The temperature that the object emits while burning
	BurnTemp = 2000,
}
local options = {
	Design = "Blizzard",
	DesignOptions = {"Blizzard", "TownBurn", "Messy",},
	BarrelDensityOptions = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,},
	BarrelDensity = 0,
	ShowHeat = true,
	ShowHeatOptions = {true, false},
}
do--Initialize
	function Fire:Initialize(width, height)
		local properties = {
			"Empty",
			"Health",
			"Temperature",
			"NewTemperature",
			"IsTree",
			"LastTurn",
		}
		grid = Grid:New(width, height, properties)

		for x, y in grid:Iterate() do
			grid:Set(x, y, "Empty", true)
			grid:Set(x, y, "Health", 100)
			grid:Set(x, y, "Temperature", 0)
			grid:Set(x, y, "NewTemperature", 0)
			grid:Set(x, y, "LastTurn", false)
			grid:Set(x, y, "IsTree", true)
		end

		for x, y in Designs:Iterate(options.Design) do
			grid:Set(x, y, "Empty", false)
		end


		
		if options.BarrelDensity > 0 then
			local density = options.BarrelDensity / 100
			for x, y in grid:Iterate() do
				if math.random() < density then
					grid:Set(x, y, "Empty", false)
					grid:Set(x, y, "IsTree", false)
				end
			end
		end
	end
	--Add to the menu system
	Menu:RegisterSimulator("Fire", Fire, options)
end

--Returns the description and mouse controls
function Fire:GetDetails()
	return "A simulation of fire spread. + is a tree and ! is an explosive barrel.", 
		"Restore a tree", 
		"Ignite a tree!"
end

do--Update
	local currentTurn = false
	--We can avoid double buffering by storing the new temp
	--and updating it to the new value as we iterate through
	local function UpdateTurn(x, y)
		if grid(x, y, "LastTurn") ~= currentTurn then
			grid:Set(x, y, "LastTurn", currentTurn)
			grid:Set(x, y, "Temperature", grid(x, y, "NewTemperature"))
		end
	end
	--Clamps a value between two bounds
	local function Clamp(value, max, min)
		if value > max then
			return max
		end
		if value < min then
			return min
		end
		return value
	end
	--Linear interpolation between values
	local function Lerp(value1, value2, amount)
		return value1 + ((value2 - value1) * amount)
	end
	local function SimulateFire(dt)
		--Usually it's best to declare the locals in smallest scope
		--However, with high iterations it's more efficient outside of the loop
		local flow, temp, burn
		local cellTemp, neighTemp
		local materialProperties
		for x, y in grid:Iterate() do
			cellTemp = grid(x, y, "Temperature")
			--Roughly simulate heat rising and leaving our system
			grid:Set(x, y, "NewTemperature", Lerp(grid(x, y, "NewTemperature"), 0, dt * .25))
			UpdateTurn(x, y)
			--Standard heat flow through the cells
			for nx, ny in grid:IterateNeighbors(x, y) do
				UpdateTurn(nx, ny)
				neighTemp = grid(nx, ny, "Temperature")

				flow = (cellTemp - neighTemp) * dt
				flow = Clamp(flow * .5, cellTemp / 4, -neighTemp / 4)

				grid:Set(x, y, "NewTemperature", grid(x, y, "NewTemperature") - flow)
				grid:Set(nx, ny, "NewTemperature", grid(nx, ny, "NewTemperature") + flow)
			end
			--Burning or potentially burning cells
			if not grid(x, y, "Empty") and grid(x, y, "Health") > 0 then
				if grid(x, y, "IsTree") then
					materialProperties = TreeProperties
				else
					materialProperties = BarrelProperties
				end
				temp = cellTemp - materialProperties.FlashPoint
				--If burning
				if temp > 0 then
					--Clamp to max burn
					if temp > materialProperties.MaxBurn * 2 then
						burn = materialProperties.MaxBurn
					--Otherwise increase the temp quadratically
					else
						burn = (1 - (.25 * temp / materialProperties.MaxBurn)) * temp
					end
					grid:Set(x, y, "NewTemperature", grid(x, y, "NewTemperature") + burn * materialProperties.BurnTemp * dt)
					grid:Set(x, y, "Health", grid(x, y, "Health") - temp * materialProperties.BurnRate)
				end
			end
		end
	end
	local function HandleInput()
		--Paint full health to a cell and its neighbors
		if IsLeftMousePressed() then
			local x, y = GetCursorPosition()
			grid:Set(x, y, "Health", 100)
			for nx, ny in grid:IterateNeighbors(x, y) do
				grid:Set(nx, ny, "Health", 100)
			end
		end
		--Burn baby burn!
		if IsRightMousePressed() then
			local x, y = GetCursorPosition()
			if grid(x, y, "NewTemperature") < TreeProperties.FlashPoint then
				grid:Set(x, y, "NewTemperature", TreeProperties.FlashPoint * 2)
			end
		end
	end
	function Fire:Update(dt)
		HandleInput()
		SimulateFire(dt)
		currentTurn = not currentTurn
	end
end

do--Drawing
	local function GetHealthColor(x, y)
		if grid(x, y, "IsTree") then
			local health = grid(x, y, "Health")
			--Healthy trees are green
			if health == 100 then
				return Colors.DarkGreen
			--Burning trees
			elseif health > 0 then
				return Colors.Red
			end
			--Dead trees
			return Colors.Gray
		else
			--Barrel
			return grid(x, y, "Health") == 100 and Colors.Yellow or Colors.Silver
		end
	end
	local function GetTemperatureColor(x, y)
		--Heat map colors
		if not options.ShowHeat then
			return Colors.Black
		end
		local temp = grid(x, y, "Temperature")
		if temp > 10 then
			return Colors.DarkRed
		elseif temp > 8 then
			return Colors.Yellow
		elseif temp > 6 then
			return Colors.Gold
		elseif temp > 4 then
			return Colors.DarkPurple
		elseif temp > 2 then
			return Colors.Teal
		end
		return Colors.Black
	end
	local function GetSymbol(x, y)
		if grid(x, y, "Empty") then
			return ' '
		end
		return grid(x, y, "IsTree") and '+' or '!'
	end
	function Fire:Draw()
		for x, y in grid:Iterate() do
			DrawSquare(x, y, GetSymbol(x, y), GetHealthColor(x, y), GetTemperatureColor(x, y))
		end
	end
end