--[[********************************************************************
created:	2009/10/10
filename: 	Menu.lua
author:		Daniel Wellman

purpose:	Provides a menu for simulators and their options

License: This software may be reproduced and distributed free of charge as
as long as credit is given to the author. The author assumes no
responsibility. The software is provided as is.
*********************************************************************]]

Menu = {}
local Menu = Menu

--Constants
local SIMULATOR_LOCATION = {X = 2, Y = 9, Delta = 2, OptionsOffset = 3}
local DETAILS_LOCATION = {X = 2, Y = 19}

--Forward Declares
local banners = {}
local buttons = {}


function Menu:Initialize()
end

function Menu:Update(delta)
	for i, banner in ipairs(banners) do
		banner:Update(delta)
	end
	for i, button in ipairs(buttons) do
		button:Update()
	end
end

--Writes a string to the buffer
local function WriteString(x, y, string, fg, bg)
	for i=1, #string do
		local index = x + i - 1
		DrawSquare(index, y, string:sub(i, i), fg, bg)
	end
end

--Draws the details of a simulator
local function DrawDetails(sim)
	local x, y = DETAILS_LOCATION.X, DETAILS_LOCATION.Y
	local info, left, right = sim:GetDetails()
	local labelColor = Colors.Purple
	local detailsColor = Colors.Green
	WriteString(x, y, "Details:", labelColor, Colors.Black)
	WriteString(x + 9, y, info, detailsColor, Colors.Black)

	WriteString(x, y + 1, "Left mouse:", labelColor, Colors.Black)
	WriteString(x + 12, y + 1, left, detailsColor, Colors.Black)

	WriteString(x, y + 2, "Right mouse:", labelColor, Colors.Black)
	WriteString(x + 13, y + 2, right, detailsColor, Colors.Black)

	WriteString(x, y + 3, "Enter:", labelColor, Colors.Black)
	WriteString(x + 7, y + 3, "Restarts the simulator", detailsColor, Colors.Black)

	WriteString(x, y + 4, "Escape:", labelColor, Colors.Black)
	WriteString(x + 8, y + 4, "Returns to this menu", detailsColor, Colors.Black)
end

local mouseOverSim
function Menu:Draw()
	--Clear the screen
	Clear()
	for i, banner in ipairs(banners) do
		banner:Draw()
	end
	for i, button in ipairs(buttons) do
		button:Draw()
	end
	if mouseOverSim then
		DrawDetails(mouseOverSim)
	end
end

local CreateButton
do
	--Returns the current index of an option from all possible options
	local function GetIndex(option, allOptions)
		for i=1, #allOptions do
			if allOptions[i] == option then
				return i
			end
		end
	end
	--Returns the longest option in the set
	local function GetLongest(allOptions)
		local longest = allOptions[1]
		for i=2, #allOptions do
			if #tostring(allOptions[i]) > #tostring(longest) then
				longest = allOptions[i]
			end
		end
		return #tostring(longest)
	end
	--Returns a formatted label
	local function GetLabel(paddingAmount, name)
		local leftPadding = string.rep(" ", math.floor((paddingAmount  - #name) / 2))
		local rightPadding = string.rep(" ", math.ceil((paddingAmount  - #name) / 2))
		return string.format("< %s%s%s >", leftPadding, name, rightPadding)
	end
	--Creates buttons for each option in the simulator
	local function CreateOptions(x, y, opts)
		for name, option in pairs(opts) do
			--Only care about non-table options
			if type(option) ~= "table" then
				--Get the full options table
				local allOptions = opts[name .. "Options"]
				local longest = GetLongest(allOptions)

				do--Label
					local label = CreateButton()
					label:SetPosition(x, y)
					label:SetText(name .. ":")
					label:SetColor(Colors.Yellow)
				end
				do--Option
					local b = CreateButton()
					b:SetPosition(x + #name + 1, y)
					b:SetText(GetLabel(longest, tostring(option)))
					b:SetColor(Colors.Purple)

					local index = GetIndex(option, allOptions)
					
					--Rotates to the next option when clicked, or back to the start if its out of options
					b:SetCallback("OnClick", function(self)
						index = index + 1
						if index > #allOptions then
							index = 1
						end
						opts[name] = allOptions[index]
						self:SetText(GetLabel(longest, tostring(opts[name])))
					end)

					b:SetCallback("OnEnter", function(self)
					    self:SetColor(Colors.Teal)
					end)
					
					b:SetCallback("OnLeave", function(self)
					    self:SetColor(Colors.Purple)
					end)
				end
				
				x = x + longest + #name + 6
			end
		end
	end
	--Keeps track of the Y location of the last simulator, so they don't overlap
	local lastY = SIMULATOR_LOCATION.Y
	function Menu:RegisterSimulator(name, sim, opts)
		local b = CreateButton()
		b:SetPosition(SIMULATOR_LOCATION.X, lastY)
		b:SetText(name)
		b:SetColor(Colors.Green)

		b:SetCallback("OnEnter", function(self)
			mouseOverSim = sim
			self:SetColor(Colors.Teal)
		end)

		b:SetCallback("OnLeave", function(self)
			mouseOverSim = sim
			self:SetColor(Colors.Green)
		end)

		b:SetCallback("OnClick", function(self)
			SetSimulator(sim)
		end)

		if opts then
			CreateOptions(SIMULATOR_LOCATION.X + SIMULATOR_LOCATION.OptionsOffset + #name, lastY, opts)
		end
		lastY = lastY + SIMULATOR_LOCATION.Delta
	end
end

do--Creates a static or moving banner
	local function Round(num)
		return math.floor(num + .5)
	end
	local function Draw(self)
		WriteString(Round(self.X), Round(self.Y), self.Message, self.ForeColor, Colors.Black)
	end
	--Scrolls the banner at the indicated speed
	local function ScrollingUpdate(self, delta)
		if not self.XSpeed then return end

		self.X = self.X + (self.UpX and self.XSpeed or -self.XSpeed) * delta
		if self.X > (WIDTH - #self.Message) then 
			self.UpX = not self.UpX
			self.X = WIDTH - #self.Message
		elseif 	self.X < 1 then
			self.UpX = not self.UpX
			self.X = 1
		end
	end
	function CreateBanner(x, y, message, color, xSpeed)
		banners[#banners + 1] = {
			Message = message,
			X = x,
			Y = y,
			UpX = true,
			XSpeed = xSpeed,
			ForeColor = color or Colors.White,
			Update = ScrollingUpdate,
			Draw = Draw,
		}
	end

	CreateBanner(1, 1, "Welcome Cellular Automata with Lua", Colors.White, 5)
	CreateBanner(10, 4, "Please select an option with the mouse, or press escape to exit")
	CreateBanner(2, 7, "Simulator             Options")
end



do	--Creates a high level button widget
	local Button = {}
	function CreateButton()
		local b = {Callbacks = {}, Color = Colors.White, Text = "", X = 1, Y = 1}

		for name, func in pairs(Button) do
			b[name] = func
		end
		buttons[#buttons + 1] = b

		return b
	end
	--Interally calls an event callback if it exists
	local function CallEvent(button, event, ...)
		return button.Callbacks[event] and button.Callbacks[event](button, ...)
	end
	
	--Sets (or removes) the function call back for an event
	function Button:SetCallback(event, callback)
		self.Callbacks[event] = callback
	end

	function Button:SetText(text)
		self.Text = tostring(text)
	end
	
	function Button:SetColor(color)
	    self.Color = color
	end
		function Button:SetPosition(x, y)
		self.X = x
		self.Y = y
	end

	--If the mouse cursor is along the length of the button its over it
	local function IsMouseOver(x, y, length)
		local mx, my = GetCursorPosition()
		return my == y and mx >= x and mx <= x + length
	end

	function Button:Update()
		--Check for mouse events
		if IsMouseOver(self.X, self.Y, #self.Text) then
			if not self.MouseIsOver then
				CallEvent(self, "OnEnter")
				self.MouseIsOver = true
			end
			if IsLeftMouseClicked() then
				CallEvent(self, "OnClick")
			end
		elseif self.MouseIsOver then
			CallEvent(self, "OnLeave")
			self.MouseIsOver = false
		end
	end
	function Button:Draw()
		WriteString(self.X, self.Y, self.Text, self.Color, Colors.Black)
	end
end