#include "Console.h"
#include <Windows.h>
#include "ConsoleBuffer.h"
#include <utility>

namespace LuaAutomata
{
	namespace
	{
		BOOL WINAPI HandlerRoutine(DWORD ctrlType)
		{
			switch(ctrlType)
			{
			case CTRL_CLOSE_EVENT:
				return true;
			}
			return false;
		}
	}

	Console::Console(unsigned width, unsigned height) 
		: m_height(height), 
		m_width(width), 
		m_frontBuffer(0), 
		m_backBuffer(0),
		m_isLeftMouseDown(false), m_isRightMouseDown(false), 
		m_isEnterDown(false), m_isEscapeDown(false),
		m_wasEscapeDown(false), m_wasEnterDown(false),
		m_wasLeftMouseDown(false),
		m_mouseX(0), m_mouseY(0)
		
	{
		SetConsoleMode(GetStdHandle(STD_OUTPUT_HANDLE), ENABLE_MOUSE_INPUT | ENABLE_PROCESSED_INPUT);

		SetConsoleCtrlHandler(HandlerRoutine, true);

		m_frontBuffer = new ConsoleBuffer(m_width, m_height);
		m_backBuffer = new ConsoleBuffer(m_width, m_height);
	}

	Console::~Console()
	{
		delete m_frontBuffer;
		delete m_backBuffer;
	}

	void Console::Update()
	{
		m_wasLeftMouseDown = m_isLeftMouseDown;
		m_wasEscapeDown = m_isEscapeDown;
		m_wasEnterDown = m_isEnterDown;
		const HANDLE input = GetStdHandle(STD_INPUT_HANDLE);

		unsigned long numEvents, unused = 0;
		GetNumberOfConsoleInputEvents(input, &numEvents);
		if (numEvents == 0)
			return;
		INPUT_RECORD* const eventBuffer = new INPUT_RECORD[numEvents];

		ReadConsoleInput(input, eventBuffer, numEvents, &unused);

		//Handle each event
		for(unsigned i = 0; i < numEvents; ++i)
		{
			switch (eventBuffer[i].EventType)
			{
			case MOUSE_EVENT:
				if (eventBuffer[i].Event.MouseEvent.dwEventFlags & MOUSE_MOVED)
				{
					m_mouseX = eventBuffer[i].Event.MouseEvent.dwMousePosition.X;
					m_mouseY = eventBuffer[i].Event.MouseEvent.dwMousePosition.Y;
				}

				m_isLeftMouseDown = (eventBuffer[i].Event.MouseEvent.dwButtonState & FROM_LEFT_1ST_BUTTON_PRESSED) != 0;
				m_isRightMouseDown = (eventBuffer[i].Event.MouseEvent.dwButtonState & RIGHTMOST_BUTTON_PRESSED) != 0;
				break;
			case KEY_EVENT:
				switch (eventBuffer[i].Event.KeyEvent.wVirtualKeyCode)
				{
				case VK_ESCAPE:
					m_isEscapeDown = eventBuffer[i].Event.KeyEvent.bKeyDown != 0;
					break;
				case VK_RETURN:
					m_isEnterDown = eventBuffer[i].Event.KeyEvent.bKeyDown != 0;
					break;
				}
				break;
			}
		}
		delete [] eventBuffer;
	}

	void Console::Draw()
	{
		//Set the front active
		ConsoleBuffer* const temp = m_frontBuffer;
		temp->SetActive();

		//Swap buffers
		m_frontBuffer = m_backBuffer;
		m_backBuffer = temp;
	}

	void Console::Clear()
	{
		m_frontBuffer->Clear();
	}

	void Console::WriteChar(char charToWrite, unsigned x, unsigned y, Color foreground, Color background)
	{
		m_frontBuffer->WriteChar(charToWrite, x, y, foreground, background);
	}

	bool Console::IsLeftMouseDown() const
	{
		return m_isLeftMouseDown;
	}

	bool Console::IsLeftMouseClicked() const
	{
		return m_wasLeftMouseDown && !m_isLeftMouseDown;
	}

	bool Console::IsRightMouseDown() const
	{
		return m_isRightMouseDown;
	}

	bool Console::IsEscapeClicked() const
	{
		return m_isEscapeDown && !m_wasEscapeDown;
	}

	bool Console::IsEnterClicked() const
	{
		return m_isEnterDown && !m_wasEnterDown;
	}

	const std::pair<unsigned, unsigned> Console::GetCursorPosition() const
	{
		return std::make_pair(m_mouseX, m_mouseY);
	}
}