#include "LuaAutomataLib.h"
#include "lua.hpp"
#include "Color.h"
#include <utility>
#include "Console.h"
#include <Windows.h>

namespace
{
	//Returns the Console from the upvalue bound to the function in AddMethod
	LuaAutomata::Console& GetConsole(lua_State* const L)
	{
		return *static_cast<LuaAutomata::Console*>(lua_touserdata(L, lua_upvalueindex(1)));
	}

	//DrawSquare(x, y, char, foreColor, backColor)
	int DrawSquare(lua_State* const L)
	{
		const int x = luaL_checkint(L, 1);
		const int y = luaL_checkint(L, 2);
		const char* const c = luaL_checkstring(L, 3);
		const LuaAutomata::Color fgColor = static_cast<LuaAutomata::Color>(luaL_checkint(L, 4));
		const LuaAutomata::Color bgColor = static_cast<LuaAutomata::Color>(luaL_checkint(L, 5));

		LuaAutomata::Console& console = GetConsole(L);

		console.WriteChar(c[0], x - 1, y - 1, fgColor, bgColor);

		return 0;
	}

	//WriteDebug("debug message")
	int WriteDebug(lua_State* const L)
	{
		const char* const line = luaL_checkstring(L, 1);
		OutputDebugStringA(line);
		OutputDebugStringA("\n");
		return 0;
	}

	//local pressed = IsLeftMousePressed()
	int IsLeftMousePressed(lua_State* const L)
	{
		lua_pushboolean(L, GetConsole(L).IsLeftMouseDown());
		return 1;
	}

	//local clicked = WasLeftMouseClicked()
	int IsLeftMouseClicked(lua_State* const L)
	{
		lua_pushboolean(L, GetConsole(L).IsLeftMouseClicked());
		return 1;
	}

	//local pressed = IsRightMousePressed()
	int IsRightMousePressed(lua_State* const L)
	{
		lua_pushboolean(L, GetConsole(L).IsRightMouseDown());
		return 1;
	}

	//local pressed = IsEnterPressed()
	int IsEnterPressed(lua_State* const L)
	{
		lua_pushboolean(L, GetConsole(L).IsEnterClicked());
		return 1;
	}

	//local pressed = IsEscapePressed()
	int IsEscapePressed(lua_State* const L)
	{
		lua_pushboolean(L, GetConsole(L).IsEscapeClicked());
		return 1;
	}

	//local x, y = GetCursorPosition()
	int GetCursorPosition(lua_State* const L)
	{
		const std::pair<int, int> pos = GetConsole(L).GetCursorPosition();
		lua_pushinteger(L, pos.first + 1);
		lua_pushinteger(L, pos.second + 1);
		return 2;
	}

	//LoadFile("filePath")
	int LoadFile(lua_State* const L)
	{
		const char* const fileName = luaL_checkstring(L, 1);
		if (luaL_loadfile(L, fileName) != 0 || lua_pcall(L, 0, 0, 0) != 0)
			luaL_argerror(L, 1, lua_tostring(L, -1));
		return 0;
	}

	//Clear()
	int Clear(lua_State* const L)
	{
		GetConsole(L).Clear();
		return 0;
	}

	template<typename ArrayType, int ArrayLength>
	void AssignColorTable(lua_State* const L, const ArrayType (&colors)[ArrayLength])
	{
		//Create a new table for the colors
		lua_createtable(L, 0, ArrayLength);

		//Fill out the table with our fields
		for (unsigned i = 0; i < ArrayLength; ++i)
		{
			//Name of color
			lua_pushstring(L, colors[i]);
			//Color index
			lua_pushnumber(L, i);
			//Set field, Colors[ColorName] = i
			lua_settable(L, -3);
		}

		//Add the table to lua
		lua_setglobal(L, "Colors");
	}

	void LoadColorTable(lua_State* const L)
	{
		const char* const colors[] = {
			"Black",
			"DarkBlue",
			"DarkGreen",
			"DarkTeal",
			"DarkRed",
			"DarkPurple",
			"Gold",
			"Gray",
			"Silver",
			"Blue",
			"Green",
			"Teal",
			"Red",
			"Purple",
			"Yellow",
			"White",
		};

		AssignColorTable(L, colors);
	}
}

namespace LuaAutomata
{
	void LuaAutomataLib::AddMethods(lua_State* const L, Console& console)
	{
		const struct luaL_reg methods[] = {
			{"LoadFile", LoadFile},
			{"GetCursorPosition", GetCursorPosition},
			{"IsEscapePressed", IsEscapePressed},
			{"IsEnterPressed", IsEnterPressed},
			{"IsRightMousePressed", IsRightMousePressed},
			{"IsLeftMousePressed", IsLeftMousePressed},
			{"WriteDebug", WriteDebug},
			{"DrawSquare", DrawSquare},
			{"Clear", Clear},
			{"IsLeftMouseClicked", IsLeftMouseClicked},
			{0, 0},
		};

		for (unsigned i = 0; methods[i].func; ++i)
		{
			lua_pushlightuserdata(L, &console);
			lua_pushcclosure(L, methods[i].func, 1);
			lua_setfield(L, LUA_GLOBALSINDEX, methods[i].name);
		}

		LoadColorTable(L);
	}
}